import pytest
from thermostat_nw.equipment_type import (
    has_heating,
    has_cooling,
    has_auxiliary,
    has_emergency,
    has_two_stage_heating,
    has_two_stage_cooling,
    has_resistance_heat,
    validate_heat_type,
    validate_cool_type,
    validate_heat_stage,
    validate_cool_stage,
    first_stage_capacity_ratio,
)


def test_missing_or_blank_heat_type():
    assert has_heating(None) is False
    assert has_heating("") is False


def test_missing_or_blank_cool_type():
    assert has_cooling(None) is False
    assert has_cooling("") is False


def test_missing_or_blank_heat_stage():
    assert has_two_stage_heating(None) is False
    assert has_two_stage_heating("") is False


def test_missing_or_blank_cool_stage():
    assert has_two_stage_cooling(None) is False
    assert has_two_stage_cooling("") is False


def test_has_heating():
    for i in [
        "furnace_or_boiler",  # Non heat pump heating (gas or oil furnace, electric resistance)
        "heat_pump_electric_backup",  # Heat pump with electric resistance heat (strip heat)
        "heat_pump_no_electric_backup",  # Heat pump without electric resistance heat
    ]:
        assert has_heating(i) is True
    for i in [
        "heat_pump_dual_fuel",  # Dual fuel heat pump (e.g. gas or oil fired)
        "other",  # Multi-zone, ?
        "none",  # None
    ]:
        assert has_heating(i) is False
    assert has_heating(None) is False
    assert validate_heat_type("furnace_or_boiler") is True
    assert validate_heat_type("bogus_heat_pump") is False


def test_has_cooling():
    for i in [
        "heat_pump",  # Heat pump w/ cooling
        "central",  # Central AC
    ]:
        assert has_cooling(i) is True
    assert has_cooling("none") is False
    assert has_cooling("other") is False
    assert has_cooling(None) is False
    assert validate_cool_type("heat_pump") is True
    assert validate_cool_type("bogus_pump") is False


def test_has_two_stage_cooling():
    assert has_two_stage_cooling("two_speed") is True
    assert has_two_stage_cooling("two_stage") is True
    assert has_two_stage_cooling("single_speed") is False
    assert has_two_stage_cooling("single_stage") is False
    assert has_two_stage_cooling("modulating") is False
    assert has_two_stage_cooling(None) is False
    assert validate_cool_stage("two_speed") is True
    assert validate_cool_stage("ten_speed") is False


def test_has_two_stage_heating():
    assert has_two_stage_heating("two_stage") is True
    assert has_two_stage_heating("single_stage") is False
    assert has_two_stage_heating("two_speed") is True
    assert has_two_stage_heating("single_speed") is False
    assert has_two_stage_heating("modulating") is False
    assert has_two_stage_heating(None) is False
    assert validate_heat_stage("two_stage") is True
    assert validate_heat_stage("two_speed") is True
    assert validate_heat_stage("2_stage") is False


def test_has_emergency():
    assert has_emergency("heat_pump_electric_backup") is True
    assert has_emergency("heat_pump_no_electric_backup") is False
    assert has_emergency("heat_pump_dual_fuel") is False


def test_has_auxiliary():
    assert has_auxiliary("heat_pump_electric_backup") is True
    assert has_auxiliary("heat_pump_no_electric_backup") is False
    assert has_auxiliary("heat_pump_dual_fuel") is False


def test_has_resistance_heat():
    assert has_resistance_heat("heat_pump_electric_backup") is True
    assert has_resistance_heat("heat_pump_no_electric_backup") is False
    assert has_resistance_heat("heat_pump_dual_fuel") is False


def test_ratio():
    assert first_stage_capacity_ratio("furnace_or_boiler") == 0.65
    assert first_stage_capacity_ratio("heat_pump") == 0.72



from setuptools import setup, find_packages, Command

version = __import__("thermostat_nw").get_version()

long_description = """
This is a fork of EPA's Connected Thermostat library (https://github.com/EPAENERGYSTAR/epathermostat), 
intended to calculate connected thermostat temperature and run-time savings. This version adds 
several metrics that are being used to evaluate real-world smart thermostat savings in the Northwestern USA.
"""


setup(
    name="thermostat_nw",
    version=version,
    description="Calculate connected thermostat savings",
    long_description=long_description,
    url="https://github.com/hshaban/epathermostat_nw",
    author="Empower Dataworks + Apex Analytics. Forked from https://github.com/EPAENERGYSTAR/epathermostat.",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    keywords="thermostat savings EPA",
    packages=find_packages(),
    package_data={"": ["*.csv", "*.json"]},
    install_requires=[
        "eemeter==2.5.2",
        "eeweather==0.3.24",
        "pandas==0.25.3",
        "numpy==1.19.4",
        "sqlalchemy==1.3.1",
        "zipcodes==1.1.2",
    ],
)



from thermostat_nw.core import Thermostat
from thermostat_nw.equipment_type import (
    has_heating,
    has_cooling,
    has_auxiliary,
    has_emergency,
    has_two_stage_cooling,
    has_two_stage_heating,
    has_multi_stage_cooling,
    has_multi_stage_heating,
    first_stage_capacity_ratio,
)

import pandas as pd
from thermostat_nw.stations import get_closest_station_by_zipcode

from thermostat_nw.eeweather_wrapper import get_indexed_temperatures_eeweather
from eeweather.cache import KeyValueStore
from eeweather.exceptions import ISDDataNotAvailableError
import json

import warnings
import dateutil.parser
import os
import pytz
from multiprocessing import Pool, cpu_count
from functools import partial
import logging

try:
    NUMBER_OF_CORES = len(os.sched_getaffinity(0))
except AttributeError:
    NUMBER_OF_CORES = cpu_count()
MAX_FTP_CONNECTIONS = 3
AVAILABLE_PROCESSES = min(NUMBER_OF_CORES, MAX_FTP_CONNECTIONS)


logger = logging.getLogger(__name__)
warnings.simplefilter("module", Warning)

INTERVAL_COLUMNS = {
    "datetime",
    "cool_runtime_stg1",
    "cool_runtime_stg2",
    "cool_runtime_equiv",
    "heat_runtime_stg1",
    "heat_runtime_stg2",
    "heat_runtime_equiv",
    "emergency_heat_runtime",
    "auxiliary_heat_runtime",
    "temp_in",
}


class ZCTAError(Exception):
    pass


def _prime_eeweather_cache():
    """Primes the eemeter / eeweather caches by doing a non-existent query
    This creates the cache directories sooner than if they were created
    during normal processing (which can lead to a race condition and missing
    thermostats)
    """
    sql_json = KeyValueStore()
    if sql_json.key_exists("0") is not False:
        raise Exception("eeweather cache was not properly primed. Aborting.")


def save_json_cache(index, thermostat_id, station, cache_path=None):
    """Saves the cached results from eeweather into a JSON file.

    Parameters
    ----------
    index : pd.DatetimeIndex
        hourly index used to compute the years needed.
    thermostat_id : str
        A unique identifier for the termostat (used for the filename)
    station : str
        Station ID used to retrieve the weather data.
    cache_path : str
        Directory path to save the cached data
    """

    json_cache = {}

    sqlite_json_store = KeyValueStore()
    years = index.groupby(index.year).keys()
    for year in years:
        filename = "ISD-{station}-{year}.json".format(station=station, year=year)
        json_cache[filename] = sqlite_json_store.retrieve_json(filename)

    if cache_path is None:
        directory = os.path.join(os.curdir, "epathermostat_weather_data")
    else:
        directory = os.path.normpath(cache_path)

    thermostat_filename = "{thermostat_id}.json".format(thermostat_id=thermostat_id)
    thermostat_path = os.path.join(directory, thermostat_filename)
    try:
        os.makedirs(os.path.dirname(directory), exist_ok=True)
        with open(thermostat_path, "w") as outfile:
            json.dump(json_cache, outfile)

    except Exception as e:
        warnings.warn("Unable to write JSON file: {}".format(e))


def normalize_utc_offset(utc_offset):
    """
    Normalizes the UTC offset
    Returns the UTC offset based on the string passed in.

    Parameters
    ----------
    utc_offset : str
        String representation of the UTC offset

    Returns
    -------
    datetime timdelta offset
    """
    try:
        if int(utc_offset) == 0:
            utc_offset = "+0"
        delta = dateutil.parser.parse(
            "2000-01-01T00:00:00" + str(utc_offset)
        ).tzinfo.utcoffset(None)
        return delta

    except (ValueError, TypeError, AttributeError) as e:
        raise TypeError("Invalid UTC offset: {} ({})".format(utc_offset, e))


def from_csv(
    metadata_filename, verbose=False, save_cache=False, shuffle=True, cache_path=None
):
    """
    Creates Thermostat objects from data stored in CSV files.

    Parameters
    ----------
    metadata_filename : str
        Path to a file containing the thermostat metadata.
    verbose : boolean
        Set to True to output a more detailed log of import activity.
    save_cache: boolean
        Set to True to save the cached data to a json file (based on Thermostat ID).
    shuffle: boolean
        Shuffles the thermostats to give them random ordering if desired (helps with caching).
    cache_path: str
        Directory path to save the cached data

    Returns
    -------
    thermostats : iterator over thermostat.Thermostat objects
        Thermostats imported from the given CSV input files.
    """

    _prime_eeweather_cache()

    metadata = pd.read_csv(
        metadata_filename,
        dtype={
            "thermostat_id": str,
            "heat_type": str,
            "heat_stage": str,
            "cool_type": str,
            "cool_stage": str,
            "zipcode": str,
            "utc_offset": str,
            "interval_data_filename": str,
        },
    )
    metadata.fillna("", inplace=True)

    # Shuffle the results to help alleviate cache issues
    if shuffle:
        logging.info("Metadata randomized to prevent collisions in cache.")
        metadata = metadata.sample(frac=1).reset_index(drop=True)

    p = Pool(AVAILABLE_PROCESSES)
    multiprocess_func_partial = partial(
        _multiprocess_func,
        metadata_filename=metadata_filename,
        verbose=verbose,
        save_cache=save_cache,
        cache_path=cache_path,
    )
    result_list = p.imap(multiprocess_func_partial, metadata.iterrows())
    p.close()
    p.join()

    # Bad thermostats return None so remove those.
    results = [x for x in result_list if x is not None]

    # Check for thermostats that were not loaded and log them
    metadata_thermostat_ids = set(metadata.thermostat_id)
    loaded_thermostat_ids = set([x.thermostat_id for x in results])
    missing_thermostats = metadata_thermostat_ids.difference(loaded_thermostat_ids)
    missing_thermostats_num = len(missing_thermostats)
    if missing_thermostats_num > 0:
        logging.warning(
            "Unable to load {} thermostat records because of "
            "errors. Please check the logs for the following thermostats:".format(
                missing_thermostats_num
            )
        )
        for thermostat in missing_thermostats:
            logging.warning(thermostat)

    # Convert this to an iterator to maintain compatibility
    return iter(results)


def _multiprocess_func(
    metadata, metadata_filename, verbose=False, save_cache=False, cache_path=None
):
    """This function is a partial function for multiproccessing and shares the same arguments as from_csv.
    It is not intended to be called directly."""
    i, row = metadata
    logger.info("Importing thermostat {}".format(row.thermostat_id))
    if verbose and logger.getEffectiveLevel() > logging.INFO:
        print("Importing thermostat {}".format(row.thermostat_id))

    interval_data_filename = os.path.join(
        os.path.dirname(metadata_filename), row.interval_data_filename
    )

    try:
        thermostat = get_single_thermostat(
            thermostat_id=row.thermostat_id,
            zipcode=row.zipcode,
            heat_type=row.heat_type,
            heat_stage=row.heat_stage,
            cool_type=row.cool_type,
            cool_stage=row.cool_stage,
            utc_offset=row.utc_offset,
            interval_data_filename=interval_data_filename,
            save_cache=save_cache,
            cache_path=cache_path,
        )
    except ZCTAError as e:
        # Could not locate a station for the thermostat. Warn and skip.
        warnings.warn(
            "Skipping import of thermostat (id={}) for which "
            "a sufficient source of outdoor weather data could not"
            "be located using the given ZIP code ({}). This is likely "
            "due to the discrepancy between US Postal Service ZIP "
            "codes (which do not always map well to locations) and "
            "Census Bureau ZCTAs (which usually do). Please supply "
            "a zipcode which corresponds to a US Census Bureau ZCTA."
            "\nError Message: {}".format(row.thermostat_id, row.zipcode, e)
        )
        return

    except ISDDataNotAvailableError as e:
        warnings.warn(
            "Skipping import of thermostat(id={} because the NCDC "
            "does not have data: {}".format(row.thermostat_id, e)
        )
        return

    except Exception as e:
        warnings.warn(
            "Skipping import of thermostat(id={}) because of "
            "the following error: {}".format(row.thermostat_id, e)
        )
        return

    return thermostat


def get_single_thermostat(
    thermostat_id,
    zipcode,
    heat_type,
    heat_stage,
    cool_type,
    cool_stage,
    utc_offset,
    interval_data_filename,
    save_cache=False,
    cache_path=None,
):
    """Load a single thermostat directly from an interval data file.

    Parameters
    ----------
    thermostat_id : str
        A unique identifier for the thermostat.
    zipcode : str
        The zipcode of the thermostat, e.g. `"01234"`.
    equipment_type : str
        The equipment type of the thermostat.
    utc_offset : str
        A string representing the UTC offset of the interval data, e.g. `"-0700"`.
        Could also be `"Z"` (UTC), or just `"+7"` (equivalent to `"+0700"`),
        or any other timezone format recognized by the library
        method dateutil.parser.parse.
    interval_data_filename : str
        The path to the CSV in which the interval data is stored.
    save_cache: boolean
        Set to True to save the cached data to a json file (based on Thermostat ID).
    cache_path: str
        Directory path to save the cached data

    Returns
    -------
    thermostat : thermostat.Thermostat
        The loaded thermostat object.
    """

    # load outdoor temperatures
    station = get_closest_station_by_zipcode(zipcode)

    if station is None:
        message = (
            "Could not locate a valid source of outdoor temperature "
            "data for ZIP code {}".format(zipcode)
        )
        raise ZCTAError(message)

    df = pd.read_csv(interval_data_filename)

    df.columns = map(str.lower, df.columns)
    if not INTERVAL_COLUMNS.issubset(df.columns):
        missing_columns = list(INTERVAL_COLUMNS.difference(df.columns))
        message = "Columns missing from interval data file: {}".format(missing_columns)
        raise ValueError(message)

    # load indices
    date_time = pd.to_datetime(df["datetime"])
    df["datetime"] = date_time
    # daily_index = pd.date_range(start=date_time[0], periods=date_time.shape[0] / 24, freq="D")
    hourly_index = pd.date_range(
        start=date_time[0], periods=date_time.shape[0], freq="H"
    )
    hourly_index_utc = pd.date_range(
        start=date_time[0], periods=date_time.shape[0], freq="H", tz=pytz.UTC
    )

    # raise an error if dates are not aligned
    if not all(date_time == hourly_index):
        missing_hours = set(hourly_index).difference(set(date_time))
        duplicates = list(date_time[date_time.duplicated()])
        message = (
            "Dates provided for thermostat_id={} may contain some "
            "which are out of order, missing, or duplicated.".format(thermostat_id)
        )
        if len(duplicates) > 0:
            message = message + " (Possible duplicated hours: {})".format(duplicates)
        elif len(missing_hours) > 0:
            message = message + " (Possible missing hours: {})".format(missing_hours)
        raise RuntimeError(message)

    # Export the data from the cache
    if save_cache:
        save_json_cache(hourly_index, thermostat_id, station, cache_path)

    # load hourly time series values
    temp_in = _create_series(df.temp_in, hourly_index)

    utc_offset = normalize_utc_offset(utc_offset)
    temp_out = get_indexed_temperatures_eeweather(
        station, hourly_index_utc - utc_offset
    )
    temp_out.index = hourly_index

    # load daily time series values
    auxiliary_heat_runtime, emergency_heat_runtime = _calculate_aux_emerg_runtime(
        df, thermostat_id, heat_type, heat_stage, hourly_index
    )
    cool_runtime = _calculate_cool_runtime(
        df, thermostat_id, cool_type, cool_stage, hourly_index
    )
    heat_runtime = _calculate_heat_runtime(
        df, thermostat_id, heat_type, heat_stage, hourly_index
    )

    # create thermostat instance
    thermostat = Thermostat(
        thermostat_id,
        heat_type,
        heat_stage,
        cool_type,
        cool_stage,
        zipcode,
        station,
        temp_in,
        temp_out,
        cool_runtime,
        heat_runtime,
        auxiliary_heat_runtime,
        emergency_heat_runtime,
    )
    return thermostat


def _calculate_cool_runtime(df, thermostat_id, cool_type, cool_stage, hourly_index):
    if has_cooling(cool_type):
        if (
            df.cool_runtime_stg1.gt(60).any()
            or df.cool_runtime_stg2.gt(60).any()
            or df.cool_runtime_equiv.gt(60).any()
        ):
            warnings.warn(
                "For thermostat {}, cooling runtime data was larger than 60 minutes"
                " for one or more hours, which is impossible. Please check the data file.".format(
                    thermostat_id
                )
            )
            return

        if has_multi_stage_cooling(cool_stage):
            warnings.warn("Multi-stage cooling isn't supported (yet)")
            return

        if has_two_stage_cooling(cool_stage):
            # Use cool equivalent runtime if it exists, otherwise calculate it
            if df.cool_runtime_equiv.max() > 0.0:
                cool_runtime = _create_series(df.cool_runtime_equiv, hourly_index)
            else:
                cool_runtime_both_stg = (
                    first_stage_capacity_ratio(cool_type)
                    * (df.cool_runtime_stg1 - df.cool_runtime_stg2)
                ) + df.cool_runtime_stg2
                cool_runtime = _create_series(cool_runtime_both_stg, hourly_index)
        else:
            cool_runtime = _create_series(df.cool_runtime_stg1, hourly_index)
    else:
        cool_runtime = None
    return cool_runtime


def _calculate_heat_runtime(df, thermostat_id, heat_type, heat_stage, hourly_index):
    if has_heating(heat_type):
        if (
            df.heat_runtime_stg1.gt(60).any()
            or df.heat_runtime_stg2.gt(60).any()
            or df.heat_runtime_equiv.gt(60).any()
        ):
            warnings.warn(
                "For thermostat {}, heating runtime data was larger than 60 minutes"
                " for one or more hours, which is impossible. Please check the data file.".format(
                    thermostat_id
                )
            )
            return

        if has_multi_stage_heating(heat_stage):
            warnings.warn("Multi-stage cooling isn't supported (yet)")
            return

        if has_two_stage_heating(heat_stage):
            # Use heat equivalent runtime if it exists, otherwise calculate it
            if df.heat_runtime_equiv.max() > 0.0:
                heat_runtime = _create_series(df.heat_runtime_equiv, hourly_index)
            else:
                heat_runtime_both_stg = (
                    first_stage_capacity_ratio(heat_type)
                    * (df.heat_runtime_stg1 - df.heat_runtime_stg2)
                ) + df.heat_runtime_stg2
                heat_runtime = _create_series(heat_runtime_both_stg, hourly_index)
        else:
            heat_runtime = _create_series(df.heat_runtime_stg1, hourly_index)
    else:
        heat_runtime = None
    return heat_runtime


def _calculate_aux_emerg_runtime(
    df, thermostat_id, heat_type, heat_stage, hourly_index
):
    if has_auxiliary(heat_type) and has_emergency(heat_type):
        auxiliary_heat_runtime = _create_series(df.auxiliary_heat_runtime, hourly_index)
        emergency_heat_runtime = _create_series(df.emergency_heat_runtime, hourly_index)
        if auxiliary_heat_runtime.gt(60).any():
            warnings.warn(
                "For thermostat {}, auxiliary runtime data was larger than 60 minutes"
                " for one or more hours, which is impossible. Please check the data file.".format(
                    thermostat_id
                )
            )
            return
        if emergency_heat_runtime.gt(60).any():
            warnings.warn(
                "For thermostat {}, emergency runtime data was larger than 60 minutes"
                " for one or more hours, which is impossible. Please check the data file.".format(
                    thermostat_id
                )
            )
            return
    else:
        auxiliary_heat_runtime = None
        emergency_heat_runtime = None
    return auxiliary_heat_runtime, emergency_heat_runtime


def _create_series(df, index):
    series = df
    series.index = index
    return series



RHU_COLUMNS = [
    "rhu1_00F_to_05F",
    "rhu1_05F_to_10F",
    "rhu1_10F_to_15F",
    "rhu1_15F_to_20F",
    "rhu1_20F_to_25F",
    "rhu1_25F_to_30F",
    "rhu1_30F_to_35F",
    "rhu1_35F_to_40F",
    "rhu1_40F_to_45F",
    "rhu1_45F_to_50F",
    "rhu1_50F_to_55F",
    "rhu1_55F_to_60F",
    "rhu1_30F_to_45F",
    "rhu2_00F_to_05F",
    "rhu2_05F_to_10F",
    "rhu2_10F_to_15F",
    "rhu2_15F_to_20F",
    "rhu2_20F_to_25F",
    "rhu2_25F_to_30F",
    "rhu2_30F_to_35F",
    "rhu2_35F_to_40F",
    "rhu2_40F_to_45F",
    "rhu2_45F_to_50F",
    "rhu2_50F_to_55F",
    "rhu2_55F_to_60F",
    "rhu2_30F_to_45F",
    "dnru_daily",
    "dnru_reduction_daily",
    "mu_estimate_daily",
    "sigma_estimate_daily",
    "sigmoid_model_error_daily",
    "sigmoid_integral_daily",
    "dnru_hourly",
    "dnru_reduction_hourly",
    "mu_estimate_hourly",
    "sigma_estimate_hourly",
    "sigmoid_model_error_hourly",
    "sigmoid_integral_hourly",
]

RHU2_IQFLT_COLUMNS = [
    "rhu2IQFLT_00F_to_05F",
    "rhu2IQFLT_05F_to_10F",
    "rhu2IQFLT_10F_to_15F",
    "rhu2IQFLT_15F_to_20F",
    "rhu2IQFLT_20F_to_25F",
    "rhu2IQFLT_25F_to_30F",
    "rhu2IQFLT_30F_to_35F",
    "rhu2IQFLT_35F_to_40F",
    "rhu2IQFLT_40F_to_45F",
    "rhu2IQFLT_45F_to_50F",
    "rhu2IQFLT_50F_to_55F",
    "rhu2IQFLT_55F_to_60F",
    "rhu2IQFLT_30F_to_45F",
]

REAL_OR_INTEGER_VALUED_COLUMNS_HEATING = [
    "n_days_in_inputfile_date_range",
    "n_days_both_heating_and_cooling",
    "n_days_insufficient_data",
    "n_core_heating_days",
    "baseline_percentile_core_heating_comfort_temperature",
    "regional_average_baseline_heating_comfort_temperature",
    "percent_savings_baseline_percentile",
    "avoided_daily_mean_core_day_runtime_baseline_percentile",
    "avoided_total_core_day_runtime_baseline_percentile",
    "baseline_daily_mean_core_day_runtime_baseline_percentile",
    "baseline_total_core_day_runtime_baseline_percentile",
    "_daily_mean_core_day_demand_baseline_baseline_percentile",
    "percent_savings_baseline_regional",
    "avoided_daily_mean_core_day_runtime_baseline_regional",
    "avoided_total_core_day_runtime_baseline_regional",
    "baseline_daily_mean_core_day_runtime_baseline_regional",
    "baseline_total_core_day_runtime_baseline_regional",
    "_daily_mean_core_day_demand_baseline_baseline_regional",
    "percent_savings_baseline_hourly_regional",
    "avoided_daily_mean_core_day_runtime_baseline_hourly_regional",
    "avoided_total_core_day_runtime_baseline_hourly_regional",
    "baseline_daily_mean_core_day_runtime_baseline_hourly_regional",
    "baseline_total_core_day_runtime_baseline_hourly_regional",
    "_daily_mean_core_day_demand_baseline_baseline_hourly_regional",
    "mean_demand",
    "alpha",
    "tau",
    "mean_sq_err",
    "root_mean_sq_err",
    "cv_root_mean_sq_err",
    "mean_abs_err",
    "mean_abs_pct_err",
    "nfev",
    "total_core_heating_runtime",
    "total_auxiliary_heating_core_day_runtime",
    "total_emergency_heating_core_day_runtime",
    "daily_mean_core_heating_runtime",
    "core_heating_days_mean_indoor_temperature",
    "core_heating_days_mean_outdoor_temperature",
    "core_mean_indoor_temperature",
    "core_mean_outdoor_temperature",
    "heat_gain_constant",
    "heat_loss_constant",
    "hvac_constant",
    "overall_temperature_variance",
    "weekly_temperature_variance",
    "avg_daily_cooling_runtime",
    "avg_daily_heating_runtime",
    "avg_daily_auxiliary_runtime",
    "avg_daily_emergency_runtime",
    "lm_intercept",
    "lm_intercept_se",
    "lm_main_slope",
    "lm_main_slope_se",
    "lm_secondary_slope",
    "lm_secondary_slope_se",
    "lm_cvrmse",
    "lm_rsquared",
] + RHU_COLUMNS

REAL_OR_INTEGER_VALUED_COLUMNS_COOLING = [
    "n_days_in_inputfile_date_range",
    "n_days_both_heating_and_cooling",
    "n_days_insufficient_data",
    "n_core_cooling_days",
    "baseline_percentile_core_cooling_comfort_temperature",
    "regional_average_baseline_cooling_comfort_temperature",
    "percent_savings_baseline_percentile",
    "avoided_daily_mean_core_day_runtime_baseline_percentile",
    "avoided_total_core_day_runtime_baseline_percentile",
    "baseline_daily_mean_core_day_runtime_baseline_percentile",
    "baseline_total_core_day_runtime_baseline_percentile",
    "_daily_mean_core_day_demand_baseline_baseline_percentile",
    "percent_savings_baseline_regional",
    "avoided_daily_mean_core_day_runtime_baseline_regional",
    "avoided_total_core_day_runtime_baseline_regional",
    "baseline_daily_mean_core_day_runtime_baseline_regional",
    "baseline_total_core_day_runtime_baseline_regional",
    "_daily_mean_core_day_demand_baseline_baseline_regional",
    "percent_savings_baseline_hourly_regional",
    "avoided_daily_mean_core_day_runtime_baseline_hourly_regional",
    "avoided_total_core_day_runtime_baseline_hourly_regional",
    "baseline_daily_mean_core_day_runtime_baseline_hourly_regional",
    "baseline_total_core_day_runtime_baseline_hourly_regional",
    "_daily_mean_core_day_demand_baseline_baseline_hourly_regional",
    "mean_demand",
    "alpha",
    "tau",
    "mean_sq_err",
    "root_mean_sq_err",
    "cv_root_mean_sq_err",
    "mean_abs_err",
    "mean_abs_pct_err",
    "nfev",
    "total_core_cooling_runtime",
    "daily_mean_core_cooling_runtime",
    "core_cooling_days_mean_indoor_temperature",
    "core_cooling_days_mean_outdoor_temperature",
    "core_mean_indoor_temperature",
    "core_mean_outdoor_temperature",
    "heat_gain_constant",
    "heat_loss_constant",
    "hvac_constant",
    "overall_temperature_variance",
    "weekly_temperature_variance",
    "avg_daily_cooling_runtime",
    "avg_daily_heating_runtime",
    "avg_daily_auxiliary_runtime",
    "avg_daily_emergency_runtime",
    "lm_intercept",
    "lm_intercept_se",
    "lm_main_slope",
    "lm_main_slope_se",
    "lm_secondary_slope",
    "lm_secondary_slope_se",
    "lm_cvrmse",
    "lm_rsquared",
]

REAL_OR_INTEGER_VALUED_COLUMNS_ALL = (
    [
        "n_days_in_inputfile_date_range",
        "n_days_both_heating_and_cooling",
        "n_days_insufficient_data",
        "n_core_cooling_days",
        "n_core_heating_days",
        "baseline_percentile_core_cooling_comfort_temperature",
        "baseline_percentile_core_heating_comfort_temperature",
        "regional_average_baseline_cooling_comfort_temperature",
        "regional_average_baseline_heating_comfort_temperature",
        "percent_savings_baseline_percentile",
        "avoided_daily_mean_core_day_runtime_baseline_percentile",
        "avoided_total_core_day_runtime_baseline_percentile",
        "baseline_daily_mean_core_day_runtime_baseline_percentile",
        "baseline_total_core_day_runtime_baseline_percentile",
        "_daily_mean_core_day_demand_baseline_baseline_percentile",
        "percent_savings_baseline_regional",
        "avoided_daily_mean_core_day_runtime_baseline_regional",
        "avoided_total_core_day_runtime_baseline_regional",
        "baseline_daily_mean_core_day_runtime_baseline_regional",
        "baseline_total_core_day_runtime_baseline_regional",
        "_daily_mean_core_day_demand_baseline_baseline_regional",
        "percent_savings_baseline_hourly_regional",
        "avoided_daily_mean_core_day_runtime_baseline_hourly_regional",
        "avoided_total_core_day_runtime_baseline_hourly_regional",
        "baseline_daily_mean_core_day_runtime_baseline_hourly_regional",
        "baseline_total_core_day_runtime_baseline_hourly_regional",
        "_daily_mean_core_day_demand_baseline_baseline_hourly_regional",
        "mean_demand",
        "alpha",
        "tau",
        "mean_sq_err",
        "root_mean_sq_err",
        "cv_root_mean_sq_err",
        "mean_abs_err",
        "mean_abs_pct_err",
        "nfev",
        "total_core_cooling_runtime",
        "total_core_heating_runtime",
        "total_auxiliary_heating_core_day_runtime",
        "total_emergency_heating_core_day_runtime",
        "daily_mean_core_cooling_runtime",
        "daily_mean_core_heating_runtime",
        "core_mean_indoor_temperature",
        "core_mean_outdoor_temperature",
        "heat_gain_constant",
        "heat_loss_constant",
        "hvac_constant",
        "overall_temperature_variance",
        "weekly_temperature_variance",
        "avg_daily_cooling_runtime",
        "avg_daily_heating_runtime",
        "avg_daily_auxiliary_runtime",
        "avg_daily_emergency_runtime",
        "lm_intercept",
        "lm_intercept_se",
        "lm_main_slope",
        "lm_main_slope_se",
        "lm_secondary_slope",
        "lm_secondary_slope_se",
        "lm_cvrmse",
        "lm_rsquared",
    ]
    + RHU_COLUMNS
    + RHU2_IQFLT_COLUMNS
)

EXPORT_COLUMNS = [
    "sw_version",
    "ct_identifier",
    "heat_type",
    "heat_stage",
    "cool_type",
    "cool_stage",
    "heating_or_cooling",
    "zipcode",
    "station",
    "climate_zone",
    "start_date",
    "end_date",
    "n_days_in_inputfile_date_range",
    "n_days_both_heating_and_cooling",
    "n_days_insufficient_data",
    "n_core_cooling_days",
    "n_core_heating_days",
    "baseline_percentile_core_cooling_comfort_temperature",
    "baseline_percentile_core_heating_comfort_temperature",
    "regional_average_baseline_cooling_comfort_temperature",
    "regional_average_baseline_heating_comfort_temperature",
    "percent_savings_baseline_percentile",
    "avoided_daily_mean_core_day_runtime_baseline_percentile",
    "avoided_total_core_day_runtime_baseline_percentile",
    "baseline_daily_mean_core_day_runtime_baseline_percentile",
    "baseline_total_core_day_runtime_baseline_percentile",
    "_daily_mean_core_day_demand_baseline_baseline_percentile",
    "percent_savings_baseline_regional",
    "avoided_daily_mean_core_day_runtime_baseline_regional",
    "avoided_total_core_day_runtime_baseline_regional",
    "baseline_daily_mean_core_day_runtime_baseline_regional",
    "baseline_total_core_day_runtime_baseline_regional",
    "_daily_mean_core_day_demand_baseline_baseline_regional",
    "percent_savings_baseline_hourly_regional",
    "avoided_daily_mean_core_day_runtime_baseline_hourly_regional",
    "avoided_total_core_day_runtime_baseline_hourly_regional",
    "baseline_daily_mean_core_day_runtime_baseline_hourly_regional",
    "baseline_total_core_day_runtime_baseline_hourly_regional",
    "_daily_mean_core_day_demand_baseline_baseline_hourly_regional",
    "mean_demand",
    "alpha",
    "tau",
    "mean_sq_err",
    "root_mean_sq_err",
    "cv_root_mean_sq_err",
    "mean_abs_err",
    "mean_abs_pct_err",
    "cov_x",
    "nfev",
    "mesg",
    "total_core_cooling_runtime",
    "total_core_heating_runtime",
    "total_auxiliary_heating_core_day_runtime",
    "total_emergency_heating_core_day_runtime",
    "daily_mean_core_cooling_runtime",
    "daily_mean_core_heating_runtime",
    "core_cooling_days_mean_indoor_temperature",
    "core_cooling_days_mean_outdoor_temperature",
    "core_heating_days_mean_indoor_temperature",
    "core_heating_days_mean_outdoor_temperature",
    "core_mean_indoor_temperature",
    "core_mean_outdoor_temperature",
    "heat_gain_constant",
    "heat_loss_constant",
    "hvac_constant",
    "overall_temperature_variance",
    "weekly_temperature_variance",
    "aux_exceeds_heat_runtime_daily",
    "aux_exceeds_heat_runtime_hourly",
    "avg_daily_cooling_runtime",
    "avg_daily_heating_runtime",
    "avg_daily_auxiliary_runtime",
    "avg_daily_emergency_runtime",
    "lm_intercept",
    "lm_intercept_se",
    "lm_main_slope",
    "lm_main_slope_se",
    "lm_secondary_slope",
    "lm_secondary_slope_se",
    "lm_cvrmse",
    "lm_rsquared",
] + RHU_COLUMNS

CERTIFICATION_HEADERS = [
    "product_id",
    "sw_version",
    "metric",
    "filter",
    "region",
    "statistic",
    "season",
    "value",
]



import pandas as pd
import numpy as np
from numpy.testing import assert_allclose

import tempfile

import pytest

from thermostat_nw.exporters import metrics_to_csv
from thermostat_nw.multiple import (
    multiple_thermostat_calculate_epa_field_savings_metrics,
)

from .fixtures.single_stage import (
    thermostat_type_1,
    thermostat_type_2,
    thermostat_type_3,
    thermostat_type_4,
    thermostat_type_5,
    metrics_type_1_data,
)
from thermostat_nw.columns import EXPORT_COLUMNS
import six


@pytest.fixture(scope="session")
def metrics_type_1(thermostat_type_1):
    metrics_type_1 = thermostat_type_1.calculate_epa_field_savings_metrics()
    return metrics_type_1


@pytest.fixture(scope="session")
def metrics_type_1_multiple(thermostat_type_1):
    metrics_type_1 = multiple_thermostat_calculate_epa_field_savings_metrics(
        [thermostat_type_1]
    )
    return metrics_type_1


RTOL = 1e-3
ATOL = 1e-3


def test_calculate_epa_field_savings_metrics_type_1(
    metrics_type_1, metrics_type_1_data
):
    assert len(metrics_type_1) == len(metrics_type_1_data)

    for key in metrics_type_1[0].keys():
        test_value = metrics_type_1[0][key]
        target_value = metrics_type_1_data[0][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)

    for key in metrics_type_1[1].keys():
        test_value = metrics_type_1[1][key]
        target_value = metrics_type_1_data[1][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)


def test_multiple_thermostat_calculate_epa_field_savings_metrics_type_1(
    metrics_type_1_multiple, metrics_type_1_data
):
    # Test multiprocessing thermostat code
    assert len(metrics_type_1_multiple) == len(metrics_type_1_data)

    for key in metrics_type_1_multiple[0].keys():
        test_value = metrics_type_1_multiple[0][key]
        target_value = metrics_type_1_data[0][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)

    for key in metrics_type_1_multiple[1].keys():
        test_value = metrics_type_1_multiple[1][key]
        target_value = metrics_type_1_data[1][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)


def test_calculate_epa_field_savings_metrics_type_2(thermostat_type_2):
    metrics_type_2_entire = thermostat_type_2.calculate_epa_field_savings_metrics()
    assert len(metrics_type_2_entire) == 2

    metrics_type_2_yearly = thermostat_type_2.calculate_epa_field_savings_metrics(
        core_cooling_day_set_method="year_end_to_end",
        core_heating_day_set_method="year_mid_to_mid",
    )
    assert len(metrics_type_2_yearly) == 9


def test_calculate_epa_field_savings_metrics_type_3(thermostat_type_3):
    metrics_type_3 = thermostat_type_3.calculate_epa_field_savings_metrics()
    assert len(metrics_type_3) == 2


def test_calculate_epa_field_savings_metrics_type_4(thermostat_type_4):
    metrics_type_4 = thermostat_type_4.calculate_epa_field_savings_metrics()
    assert len(metrics_type_4) == 1


def test_calculate_epa_field_savings_metrics_type_5(thermostat_type_5):
    metrics_type_5 = thermostat_type_5.calculate_epa_field_savings_metrics()
    assert len(metrics_type_5) == 1


def test_metrics_to_csv(metrics_type_1):

    fd, fname = tempfile.mkstemp()
    df = metrics_to_csv(metrics_type_1, fname)

    assert isinstance(df, pd.DataFrame)
    assert df.columns[0] == "sw_version"
    assert df.columns[1] == "ct_identifier"

    with open(fname, "r") as f:
        lines = f.readlines()
        assert len(lines) == 3
        column_heads = lines[0].strip().split(",")
        assert column_heads == EXPORT_COLUMNS
